/* -- jimbob, a specialized Markov chain object for Pluton, etc., by
Philippe Manoury.  It's not clear whom this code legally belongs to,
so caution is advised for people wanting to incorporate this module into their
own ventures.  Given the brevity of this module, such people might want
to consider simply re-writing it.  It's my belief that none of the possible
owners (IRCAM? Manoury?) will object to anyone using it for artistic
purposes.  */

#include "m_pd.h"
#include <math.h>

/*************** jimbob **************************/

#define DIM 88
#define GOO 8
#define COLS (DIM + GOO)
#define FIRST 21
#define LAST (FIRST + DIM)

typedef struct trans
{
    unsigned int t_vel;
    unsigned int t_length;
    unsigned int t_inter;
    unsigned int t_pad;
} t_trans;

typedef struct hang
{
    struct trans *h_trans;
    double h_time;
} t_hang;

typedef struct jimbob
{
    t_object m_ob;
    struct trans *m_matrix;
    int m_recpit;
    int m_restart;
    double m_rectime;
    struct hang m_hang[DIM];
    char m_goophase[DIM];
    struct trans m_firsttrans;
    int m_firstpit;
    int m_playpit;
    struct trans *m_playtrans;

    t_clock *m_clock;
    t_outlet *m_pitchout;
    t_outlet *m_velout;
    t_outlet *m_durout;
    long m_val;
    long m_maxdiff;
    long m_record;
    t_float m_inval;
    long m_repeat;
    int m_chordlength;
} t_jimbob;

void jimbob_tick(t_jimbob *m);
void jimbob_clear(t_jimbob *m);

static long jimbob_randi_state;

static int jimbob_randi(void)
{
    jimbob_randi_state = jimbob_randi_state * 210427233L + 1489352071L;
    return ((jimbob_randi_state >> 16) & 0x7fff);
}

static t_class *jimbob_class;

static void *jimbob_new(void)
{
    t_jimbob *m;
    int i, j;
    struct trans *tp;

    m = (t_jimbob *)pd_new(jimbob_class);
    m->m_matrix = (struct trans *)
	t_getbytes((sizeof(struct trans) * (long)DIM * (long) COLS));
    m->m_clock = clock_new(m, (t_method)jimbob_tick);
	floatinlet_new(&m->m_ob, &m->m_inval);
    m->m_pitchout = outlet_new(&m->m_ob, gensym("float"));
    m->m_velout = outlet_new(&m->m_ob, gensym("float"));
    m->m_durout = outlet_new(&m->m_ob, gensym("float"));
    m->m_restart = 0;
    jimbob_clear(m);
    m->m_val = 2000;
    m->m_maxdiff = 50;
    m->m_inval = 0;
    m->m_repeat = 0;
    m->m_chordlength = 0;
    return (m);
}

static void jimbob_float(t_jimbob *m, t_floatarg fx1)
{
    long x1 = fx1;
    long x2;
    int lastrec;
    long timediff;

    x2 = m->m_inval;
    lastrec = m->m_recpit;
    if (!m->m_record)
    	return;
    x1 -= FIRST;
    if (!x2 && x1 >= 0 && x1 < DIM)
    {
	struct hang *h = &m->m_hang[x1];
	if (h->h_trans) 
	{
	    	/* time stored in 8-msec units, originally to fit into one
		byte of memory.  It might be dangerous to change this now!
		-msp 99 */
	    long del = (clock_gettimesince(h->h_time)) * 0.125;
	    if (del > 255) del = 255;
	    else if (del < 1) del = 1;
	    h->h_trans->t_length = del;
	    h->h_trans = 0;
	}
	return;
    }
    if (x2 && m->m_firstpit < 0)
    {
	m->m_firstpit = x1;
	m->m_firsttrans.t_vel = x2;
	m->m_firsttrans.t_length = 64;
	m->m_hang[x1].h_time = clock_getsystime();
	m->m_hang[x1].h_trans = &m->m_firsttrans;
    }
    else if (x2 && x1 >= 0 && x1 < DIM && lastrec >= 0 && lastrec < DIM
	   && (timediff = clock_gettimesince(m->m_rectime)) < m->m_val)
    {
	struct trans *t = &(m->m_matrix)[lastrec*COLS + x1];	
	/* save up to GOO old transitions for repeated notes */
	if (x1 == lastrec)
	{
	    int phase = m->m_goophase[lastrec];
	    struct trans *t2 = &(m->m_matrix)[lastrec*COLS +
					     DIM + phase];
	    *t2 = *t;
	    m->m_goophase[lastrec] = (phase + 1) & (GOO-1);
	}
	t->t_vel = x2;
	t->t_length = 64;
	if (timediff < m->m_maxdiff) timediff = 0;
	t->t_inter = timediff;
	m->m_hang[x1].h_time = clock_getsystime();
	m->m_hang[x1].h_trans = t;
    }
    if (x2)
    {
	m->m_recpit = x1;
	m->m_rectime = clock_getsystime();
    }
}

void jimbob_print(t_jimbob *m)
{
    post("record %ld time %ld", (long)m->m_recpit, m->m_rectime);
    post("val %ld", m->m_val);
    post("maxdiff %ld", m->m_maxdiff);
}

void jimbob_clear(t_jimbob *m)
{
    int i, j;
    struct trans *tp;

    m->m_record = 0;
    for (i=0; i<DIM; i++) for (j = COLS, tp=(m->m_matrix) + i*COLS; j--; tp++)
	tp->t_vel = 0;
    for (i=0; i<DIM; i++) m->m_hang[i].h_trans = 0, m->m_goophase[i] = 0;
    m->m_playtrans = 0;
    m->m_playpit = 0;
    m->m_recpit = -1;
    m->m_firstpit = -1;
    clock_unset(m->m_clock);
}

void jimbob_record(t_jimbob *m)
{
    m->m_record = 1;
}

void jimbob_stoprec(t_jimbob *m)
{
    m->m_record = 0;
}

void jimbob_bang(t_jimbob *m)
{
    if (m->m_firstpit < 0)
    {
	post("jimbob: no sequence");
	return;
    }
    m->m_playpit = m->m_firstpit;
    m->m_playtrans = &m->m_firsttrans;
    clock_delay(m->m_clock, 0);
}

void jimbob_tick(t_jimbob *m)
{
    long x, r, index;
    int i, row, lastplay, ncolumn;
    struct trans *tp;

    x = 0;
    r = jimbob_randi();
    ncolumn = (m->m_repeat ? COLS : DIM);
    outlet_float(m->m_durout, (m->m_playtrans->t_length & 0xffL)<<3);
    outlet_float(m->m_velout, m->m_playtrans->t_vel & 0xffL);
    outlet_float(m->m_pitchout, (long)m->m_playpit + FIRST);
    row = m->m_playpit;
    for (tp = (m->m_matrix) + COLS * row, i=ncolumn; i--; tp++)
    x += tp->t_vel;
    /* post("%ld->%ld", (long)row, x); */
    if (!x)				/* if nowhere to go stop */
    {
	if (m->m_restart)
	{
	    if (m->m_firstpit < 0)
	    {
		post("jimbob: no sequence");
		return;
	    }
	    m->m_playpit = m->m_firstpit;
	    m->m_playtrans = &m->m_firsttrans;
	    clock_delay(m->m_clock, 500);
	}
	return;
    }
    index = (r * x) >> 15;
    if (!index) index++;
    for (tp = (m->m_matrix) + COLS * row, i=0; i < ncolumn; tp++, i++)
    {
	index -= tp->t_vel;
	if (index <= 0) break;
    }
    if (i >= ncolumn) post("fuck"), i--;
    m->m_playtrans = (m->m_matrix) + COLS * row + i;
    m->m_playpit = (i >= DIM ? row:i);
    if (m->m_playtrans->t_inter < 5) m->m_chordlength++;
    else m->m_chordlength = 0;
    if (m->m_chordlength > 50) clock_delay(m->m_clock, 500);
    else clock_delay(m->m_clock, m->m_playtrans->t_inter);
}

void jimbob_stop(t_jimbob *m)
{
    clock_unset(m->m_clock);
}

void jimbob_set(t_jimbob *m, t_floatarg f)
{
    m->m_val = f;
}

void jimbob_restart(t_jimbob *m, t_floatarg f)
{
    m->m_restart = f;
}

void jimbob_r(t_jimbob *m, t_floatarg f)
{
    m->m_repeat = f;
}

void jimbob_maxdiff(t_jimbob *m, t_floatarg f)
{
    m->m_maxdiff = f;
}

void jimbob_ff(t_jimbob *m)
{
    clock_free(m->m_clock);
    t_freebytes(m->m_matrix, (sizeof(struct trans) * (long)DIM * (long) COLS));
}

void jimbob_setup(void)
{
    jimbob_class = class_new(gensym("jimbob"),(t_newmethod)jimbob_new,
    	(t_method)jimbob_ff, sizeof(t_jimbob), 0, 0);
    class_addbang(jimbob_class, jimbob_bang);
    class_addfloat(jimbob_class, jimbob_float);
    class_addmethod(jimbob_class, (t_method)jimbob_print,
    	gensym("print"), 0);
    class_addmethod(jimbob_class, (t_method)jimbob_clear,
    	gensym("clear"), 0);
    class_addmethod(jimbob_class, (t_method)jimbob_record,
    	gensym("record"), 0);
    class_addmethod(jimbob_class, (t_method)jimbob_stoprec,
    	gensym("stoprec"), 0);
    class_addmethod(jimbob_class, (t_method)jimbob_stop,
    	gensym("stop"), 0);
    class_addmethod(jimbob_class, (t_method)jimbob_set,
    	gensym("set"), A_FLOAT, 0);
    class_addmethod(jimbob_class , (t_method)jimbob_maxdiff,
    	gensym("maxdiff"), A_FLOAT, 0);
    class_addmethod(jimbob_class, (t_method)jimbob_restart,
    	gensym("restart"), A_FLOAT, 0);
    class_addmethod(jimbob_class, (t_method)jimbob_r,
    	gensym("r"), A_FLOAT, 0);
}
